import argparse
import signal
import sys
from blessed import Terminal

PH = "⬚"

parser = argparse.ArgumentParser(description="Edit a file.")
parser.add_argument("filename")
args = parser.parse_args()

buffer = []

def load(f):
    content = f.read()
    if not content:
        buffer.clear()
        return
    if any(ord(c) not in {10, *range(32, 126)} for c in content):
        raise ValueError("file is not printable ASCII")
    buffer[:] = [list(x + PH*x.endswith(" ")) for x in content.split("\n")]
    if not buffer[0]:
        buffer[0].append(PH)
    if not buffer[-1]:
        buffer[-1].append(PH)

def save(f):
    for i, line in enumerate(buffer):
        end = "\n" if i + 1 != len(buffer) else ""
        f.write("".join(c for c in line if c != PH) + end)

def at(y, x):
    y += origin
    if 0 <= y < len(buffer) and 0 <= x < len(buffer[y]):
        return buffer[y][x]
    else:
        return " "

def put(y, x, c):
    global origin
    if x < 0:
        return
    if not buffer:
        origin = -y
    if c != " ":
        y += origin
        prepend_by = max(-y, 0)
        origin += prepend_by
        y += prepend_by
        buffer[:0] = [[] for _ in range(prepend_by)]
        buffer.extend([] for _ in range(max(y-len(buffer)+1, 0)))
        line = buffer[y]
        line.extend(" " * max(x-len(line)+1, 0))
        line[x] = c
    else:
        if at(y, x) == " ":
            return
        clear_off.append((y, x))
        y += origin
        line = buffer[y]
        line[x] = " "
        while line and line[-1] == " ":
            line.pop()
        while buffer and not buffer[0]:
            origin -= 1
            buffer.pop(0)
        while buffer and not buffer[-1]:
            buffer.pop()

try:
    with open(args.filename) as f:
        load(f)
except FileNotFoundError:
    pass
except ValueError:
    print("oh no: file does not consist of printable ASCII", file=sys.stderr)
    sys.exit(1)

clear_off = []
origin = -1
uy = ux = cy = cx = 0
at_keyboard = False

term = Terminal()
norm = term.bright_white + term.on_color_rgb(0x26, 0x26, 0x26)
sb = term.white + term.on_color_rgb(0x36, 0x36, 0x36)

def total_clear():
    print(norm + term.clear)
    clear_off.clear()

def render():
    global cy, cx, ux

    o = cy, cx
    h = term.height
    cy += (uy-cy) // h * h
    s = cy + origin
    nw = len(str(max(s, 0) + h))

    if not at_keyboard and ux < ~nw:
        add_keyboard()
    if at_keyboard and ux >= at_keyboard:
        remove_keyboard()
    if at_keyboard:
        nw = -1
        ux = max(ux, 0)

    w = term.width - nw - 1
    cx += (ux-cx) // w * w
    cx = max(cx, 0)

    if (cy, cx) != o:
        total_clear()

    with term.hidden_cursor():
        print(norm, end="")
        for y, x in clear_off:
            print(term.move_yx(y - cy, x - cx + nw+1) + " ", end="")
        clear_off.clear()

        print(term.home, end="")
        for dry in range(s, s+h):
            end = "\r\n" if dry + 1 != s+h else ""
            if 0 <= dry < len(buffer):
                line = "".join(buffer[dry][cx:cx+w])
                if not at_keyboard:
                    print(f"{sb}{dry+1:>{nw}} ", end="")
                print(f"{norm}{line}", end=end)
            else:
                print("" if at_keyboard else sb + " "*(nw+1), end=end)

        print(term.move_yx(uy - cy, ux - cx + nw+1), end="", flush=True)

def on_resize(sig, action):
    render()

signal.signal(signal.SIGWINCH, on_resize)

def shift(dy, dx):
    global uy, ux
    uy += dy
    ux += dx
    y = uy
    x = ux
    shifting = " "
    while at(y, x) != " ":
        temp = at(y, x)
        put(y, x, shifting)
        shifting = temp
        y += dy
        x += dx
    put(y, x, shifting)

KEYBOARD = rf"""
` ~ 1 ! 2 @ 3 # 4 $ 5 % 6 ^ 7 & 8 * 9 ( 0 ) - _ = +

     q Q w W e E r R t T y Y u U i I o O p P [ {{ ] }} \ |

      a A s S d D f F g G h H j J k K l L ; : ' "

       z Z x X c C v V b B n N m M , < . > / ? {PH}
"""

def add_keyboard():
    global at_keyboard, ux
    shift = term.width + ~ux
    at_keyboard = shift
    ux += shift
    for l in buffer:
        if l:
            l[:0] = [" "] * shift
    rows = KEYBOARD.strip().splitlines()
    top = term.height // 2 - len(rows) // 2
    left = term.width // 2 - len(rows[2]) // 2
    for y, row in enumerate(rows):
        for x, c in enumerate(row):
            put(top+y, left+x, c) 
    total_clear()

def remove_keyboard():
    global at_keyboard, ux
    for y, l in enumerate(buffer):
        for x in range(0, at_keyboard):
            put(y-origin, x, " ")
        del l[:at_keyboard]
    ux -= at_keyboard
    at_keyboard = None

def ctrl(c):
    return chr(ord(c) & 31)

with term.fullscreen(), term.raw():
    total_clear()
    render()
    while True:
        next_key = term.inkey()
        if next_key == ctrl("q"):
            break
        elif next_key == ctrl("s"):
            with open(args.filename, "w") as f:
                save(f)
        elif next_key.code == term.KEY_UP:
            shift(-1, 0)
        elif next_key.code == term.KEY_DOWN:
            shift(1, 0)
        elif next_key.code == term.KEY_RIGHT:
            shift(0, 1)
        elif next_key.code == term.KEY_LEFT:
            shift(0, -1)
        render()
